package Mentat::Processor::Engine::Briefer;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Processor::Engine::Briefer - Short module description

=head1 SYNOPSIS

 use Mentat::Processor::Engine::Briefer;

=head1 DESCRIPTION

Implementation of processing engine performing event reporting.

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Mentat::Processor::Engine(3), Mentat::Message::Grouper(3), Data::Filter(3),
Mentat::Stats::Data(3), Mentat::Storage::Mongo(3), Whois::Dump(3), Mentat::Message::Formater::TextTemplate(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Time::Local;
use IO::Compress::Zip qw(zip $ZipError);
use Text::Wrap;
use Number::Format qw(:subs :vars);
use Mail::Sendmail;
use Encode;

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use JSON;
use Time::Local;
use Text::Template;

#-- Custom application modules ------------------------------------------------#
use Log::Loger;
use Value::Convertor;
use Mentat::Core qw(:c_general :c_hawat);
use Mentat::Processor::Engine;
use Mentat::Message::Grouper;
use Data::Filter;
use Mentat::Stats::Brief;
use Mentat::Storage::Mongo;
use Whois::Service;
use Whois::Mongo;
use Whois::Negistry;
use Mentat::Message::Formater::TextTemplate;
use Mentat::Mailer::Report;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant DFLT_DATABASE         => 'mentat';
use constant DFLT_BRIEF_COLLECTION => 'briefs';

use constant LAST_RUN_TS        => 'last_run_ts';
use constant MAX_INTERVAL_DELTA => 5;

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.6';
    $DEVEL   = 0;
    @ISA     = ('Mentat::Processor::Engine');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item eh_process($) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object

=cut

sub eh_process($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my ($th, $tl, $date, $type) = $self->time_boundaries();
    Log::Loger->notice(__PACKAGE__, "[ENG:Briefer] Briefing start with time window '".localtime($tl)."' - '".localtime($th)."' [$tl - $th]");

    # Storage for general statistics
    my $runlog = {
        'ts_start' => int(time), # Timestamp of calculation beginning
        'ts_from'  => int($tl),  # Lower threshold
        'ts_to'    => int($th),  # Higher threshold
        'date'     => $date,
        'type'     => $type,
    };

    # Gather the various system statistics
    my $proc_stats = $self->{STATS}->calc_stats_processing($th, $tl);
    my $rep_stats  = $self->{STATS}->calc_stats_reporting($th, $tl);
    print STDERR "Proc: ".Dumper($proc_stats) if $DEVEL;
    print STDERR "Rep: ".Dumper($rep_stats) if $DEVEL;

    my $sections_proc = [];
    push(@$sections_proc, $self->{TEMPLATE_SPROC}->fill_in(HASH => {
                'HL'    => \$self->{HELPER},
                'LABEL' => 'TOTAL PROCESSING STATISTICS',
                'STATS' => $proc_stats->{overall},
                'COUNT' => $proc_stats->{cnt_all},
            }));
    if ($proc_stats->{internal}) {
        push(@$sections_proc, $self->{TEMPLATE_SPROC}->fill_in(HASH => {
                'HL'    => \$self->{HELPER},
                'LABEL' => 'PROCESSING STATISTICS FOR CESNET2 NETWORKS',
                'STATS' => $proc_stats->{internal},
                'COUNT' => $proc_stats->{cnt_all},
            }));
        push(@$sections_proc, $self->{TEMPLATE_SPROC}->fill_in(HASH => {
                'HL'    => \$self->{HELPER},
                'LABEL' => 'PROCESSING STATISTICS EXCLUDING CESNET2 NETWORKS',
                'STATS' => $proc_stats->{external},
                'COUNT' => $proc_stats->{cnt_all},
            }));
    }

    my $section_rep = $self->{TEMPLATE_SREP}->fill_in(HASH => {
                'HL'    => \$self->{HELPER},
                'STATS' => $rep_stats,
            });

    my $brief = $self->{TEMPLATE}->fill_in(HASH => {
            'HL'      => \$self->{HELPER},
            'TH'      => $th,
            'TL'      => $tl,
            'BD'      => $date,
            'BT'      => $type,
            'TF_P'    => $proc_stats->{ts_from},
            'TT_P'    => $proc_stats->{ts_to},
            'TF_R'    => $rep_stats->{ts_from},
            'TT_R'    => $rep_stats->{ts_to},
            'COUNT'   => $proc_stats->{cnt_all},
            'COUNT_R' => $rep_stats->{cnt_mails} || 0,
            'PSECTS'  => $sections_proc,
            'RSECTS'  => [$section_rep],
        });
    $self->_brief_email($brief, $th, $tl);

    # Set the runlog, it will be dumped automatically after the processing is done
    $runlog->{ts}       = time;
    $runlog->{duration} = $self->duration();

    $self->runlog_set($runlog);

    return (EC_CONTINUE);
}

=item time_boundaries() [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object

=cut

sub time_boundaries()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    return ($self->{TIME_HIGH}, $self->{TIME_LOW}, $self->{BRIEF_DATE}, $self->{BRIEF_TYPE});
}

#-------------------------------------------------------------------------------

=item startup(;$) [PUBLIC]

  Usage     : $instance->startup();
  Purpose   : Callback to allow the component perform arbitrary startup operations
  Arguments : HASH REFERENCE $config - startup configurations [OPTIONAL]
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance or on error

=cut

sub startup(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($config,) = @_;
    confess "Component startup configurations must be passed via HASH reference" if $config and ref $config ne 'HASH';

    # Check time boundaries
    $self->_check_timing();
}

=item pstate($;$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on object instance

=cut

sub pstate($;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->component('pstate')->pstate(@_);
}

=item pstate_set($$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on object instance

=cut

sub pstate_set($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->component('pstate')->pstate_set(@_);
}

=item pstate_del($) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on object instance

=cut

sub pstate_del($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->component('pstate')->pstate_del(@_);
}

=item pstate_list() [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on object instance

=cut

sub pstate_list()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->component('pstate')->pstate_list();
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Processor::Engine::Briefer instance
#  Arguments :
#  Returns   : Mentat::Processor::Engine::Briefer reference
#  Throws    : Dies, if not invoked on object instance
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my %config = @_;

    # Call parent version first
    $self = $self->SUPER::_init(%config);

    die "[ENG:Briefer] Template directory must be specified under configuration key 'template_dir'\n" unless exists $config{template_dir};
    die "[ENG:Briefer] Template directory '$config{'template_dir'}' does not exist\n" unless -d $config{template_dir};
    die "[ENG:Briefer] Brief type must be specified under configuration key 'brief_type'\n" unless exists $config{brief_type};
    die "[ENG:Briefer] Report target email address must be specified under configuration key 'mail_to'\n" unless $config{mail_to};
    die "[ENG:Briefer] Report origin email address must be specified under configuration key 'mail_from'\n" unless $config{mail_from};
    die "[ENG:Briefer] Report reply-to email address must be specified under configuration key 'reply_to'\n" unless $config{reply_to};
    die "[ENG:Briefer] Email address for system administrators must be specified under configuration key 'mail_admin'\n" unless $config{mail_admin};
    die "[ENG:Briefer] Email address for system developers must be specified under configuration key 'mail_devel'\n" unless $config{mail_devel};
    die "[ENG:Briefer] Report directory must be specified under configuration key 'reports_dir'\n" unless exists $config{reports_dir};
    die "[ENG:Briefer] Report directory '$config{'reports_dir'}' does not exist\n" unless -d $config{reports_dir};
    die "[ENG:Briefer] Report directory '$config{'reports_dir'}' is not writable\n" unless -w $config{reports_dir};
    die "[ENG:Briefer] Chart directory must be specified under configuration key 'charts_dir'\n" unless exists $config{charts_dir};
    die "[ENG:Briefer] Chart directory '$config{charts_dir}' does not exist\n" unless -d $config{charts_dir};
    die "[ENG:Briefer] Chart directory '$config{charts_dir}' is not writable\n" unless -w $config{charts_dir};

    $self->{TO}          = $config{mail_to};
    $self->{FROM}        = $config{mail_from};
    $self->{REPLY_TO}    = $config{reply_to};
    $self->{ADMIN_C}     = $config{mail_admin};
    $self->{TEST_C}      = $config{mail_devel};
    $self->{BCC_C}       = $config{mail_bcc};
    $self->{TEST_MODE}   = $config{test_mode};

    $self->{STATS}       = Mentat::Stats::Brief->new(%config);

    Log::Loger->info(__PACKAGE__, "[ENG:Briefer] Using email '$self->{TO}' as target email address");
    Log::Loger->info(__PACKAGE__, "[ENG:Briefer] Using email '$self->{FROM}' as source email address");
    Log::Loger->info(__PACKAGE__, "[ENG:Briefer] Using email '$self->{REPLY_TO}' as return email address");
    Log::Loger->info(__PACKAGE__, "[ENG:Briefer] Using email '$self->{ADMIN_C}' as admin contact email address");
    Log::Loger->info(__PACKAGE__, "[ENG:Briefer] Using email '$self->{TEST_C}' as developer contact email address");
    Log::Loger->info(__PACKAGE__, "[ENG:Briefer] Using email '$self->{BCC_C}' as BCC email address") if $self->{BCC_C};
    Log::Loger->info(__PACKAGE__, "[ENG:Briefer] Running in test mode, using email '$self->{TEST_C}' as destination email address for all reports") if $self->{TEST_MODE};

    $self->{TEMPLATE}       = Text::Template->new(TYPE => 'FILE', SOURCE => $config{template_dir}.'/_master.tt', UNTAINT => 1);
    $self->{TEMPLATE_SPROC} = Text::Template->new(TYPE => 'FILE', SOURCE => $config{template_dir}.'/_stats_proc.tt', UNTAINT => 1);
    $self->{TEMPLATE_SREP}  = Text::Template->new(TYPE => 'FILE', SOURCE => $config{template_dir}.'/_stats_rep.tt', UNTAINT => 1);

    $self->{HELPER}   = Mentat::Message::Formater::Helper->singleton(
                                    resolve => $config{resolve},
                                    dict    => $config{dict},
                                    plurals => {
                                            'uaddress' => ['unikátní IP adresa','unikátní IP adresy','unikátních IP adres'],
                                            'attempts' => ['pokus','pokusy','pokusů'],
                                            'events'   => ['událost','události','událostí'],
                                        },
                        );

    my ($th, $tl, $date, $type) = $self->{STATS}->calculate_thresholds($config{brief_date}, $config{brief_type});

    # Calculate time boundaries
    $self->{RUN_INTERVAL} = $th - $tl;
    $self->{TIME_HIGH}    = $th;
    $self->{TIME_LOW}     = $tl;
    $self->{BRIEF_DATE}   = $date;
    $self->{BRIEF_TYPE}   = $type;

    unless ($config{brief_date}) {
        $self->{TIME_HIGH} -= $self->{RUN_INTERVAL};
        $self->{TIME_LOW}  -= $self->{RUN_INTERVAL};
        $self->{BRIEF_DATE} = $self->{STATS}->brief_date($self->{TIME_LOW});
    }
    Log::Loger->debug(__PACKAGE__, "[ENG:Briefer] Brief for '".$self->{BRIEF_DATE}."'/'".$self->{BRIEF_TYPE}."', time boundaries: " . localtime($self->{TIME_LOW}) . " - " . localtime($self->{TIME_HIGH}) . " (".$self->{TIME_LOW}."-".$self->{TIME_HIGH}.")");

    return $self;
}

# _get_components_linked(;@) [PROTECTED]
#
#  Usage     : Used as a hook method from template method register()
#  Purpose   : Return list of required components
#  Arguments : LIST OF ARRAY REFERENCES @lists - linked component lists from parent classess [OPTIONAL]
#  Returns   : ARRAY REFERENCE $components containing STRINGS with following format:
#                  'type'      - request for component of given type and any name
#                  ':name'     - request for component of any type but given name
#                  'type:name' - request for component of given both type and name
#  Throws    : Dies, if not invoked on instance

sub _get_components_linked(;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my @lists = @_;

    return $self->SUPER::_get_components_linked([
        ['state', 'pstate'],
    ], @lists);
}

# _check_timing() [PROTECTED]
#
#  Usage     : $self->_check_timing();
#  Purpose   : Check procesing time boundaries
#  Arguments : Hash containing application configurations
#  Returns   :
#  Throws    :
#  Comments  :
#  See Also  :

sub _check_timing()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my $current_time = time();

    # Retrieve the last and save the current LAST RUN TIME
    $self->{LAST_RUN_TS} = $self->pstate(LAST_RUN_TS());
    $self->pstate_set(LAST_RUN_TS(), $current_time);

    # Check the proper interval
    if ($self->{LAST_RUN_TS}) {
        $self->{RUN_INTERVAL_ACT}   = $current_time - $self->{LAST_RUN_TS};
        $self->{RUN_INTERVAL_DELTA} = $self->{RUN_INTERVAL_ACT} - $self->{RUN_INTERVAL};

        if (abs($self->{RUN_INTERVAL_DELTA}) > MAX_INTERVAL_DELTA()) {
            my $how = ($self->{RUN_INTERVAL_DELTA} < 0) ? 'soon' : 'late';
            Log::Loger->warn(__PACKAGE__, "[ENG:Briefer] Running too $how with interval ".$self->{RUN_INTERVAL_ACT}." seconds after last run at ".$self->{LAST_RUN_TS}." (".abs($self->{RUN_INTERVAL_DELTA})." seconds $how)");
        }
    }
}

#-------------------------------------------------------------------------------

sub _brief_email($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($brief, $th, $tl) = @_;

    my $to = $self->{TO};

    # Test mode beats everything, redirect to system developers
    if ($self->{TEST_MODE}) {
        Log::Loger->warn(__PACKAGE__, "[ENG:Briefer] Running in test mode, redirecting system briefing from <$to> to system developers at <$self->{TEST_C}>");
        $brief = "WARNING: Running in test mode, system briefing was redirected from <$to> to system developers at <$self->{TEST_C}>.\n\n$brief";
        $to = $self->{TEST_C};
    }


    my %mail = (
            'From'         => $self->{FROM},
            'To'           => $to,
            'Reply-To'     => $self->{REPLY_TO},
        );
    $mail{'Reply-To'} = $self->{REPLY_TO} if $self->{REPLY_TO};
    $mail{'Bcc'}      = $self->{BCC_C}    if $self->{BCC_C};

    my $tp = $self->{HELPER}->time($tl, 1).' - '.$self->{HELPER}->time($th, 1);
    # Encode the email subject
    {
        use utf8;
        $mail{Subject} = encode("MIME-Header", "Mentat system performance statistics ($tp)");
    }

    $mail{'Content-Type'} = 'text/plain; charset="utf-8"';
    $mail{'Message'} = "$brief\n";

    # Send the email
    Log::Loger->info(__PACKAGE__, "[ENG:Briefer] Attempting to mail briefing to <$to>");
    my $success = sendmail(%mail);
    if ($success) {
        my $result;
        if ($Mail::Sendmail::log =~ /(Result: [^\n]+)/) { $result = $1; }
        Log::Loger->notice(__PACKAGE__, "[ENG:Briefer] Mailed briefing to <$to>: " . $result);
        #Log::Loger->notice(__PACKAGE__, "[ENG:Briefer] Mailed briefing to <$to>: " . $Mail::Sendmail::log);
        return ($success, $result, $to);
    }
    else {
        Log::Loger->error(__PACKAGE__, "[ENG:Briefer] Unable to mail briefing to <$to>: " . $Mail::Sendmail::error);
        return ($success, $Mail::Sendmail::error);
    }
}

=pod

=back

=cut

1;
