package Whois::Service;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Whois::Service - Simple whois database constructed by parsing whois dump file

=head1 SYNOPSIS

 use Whois::Service;

 # Instantinate the whois database by parsing given whois dump, file must not be
 # older than three days
 my $wi_tree = Whois::Service->new('./data/whois.dump');

 # Instantinate the whois database by parsing given whois dump, do not check the
 # file age
 my $wi_tree = Whois::Service->new('./data/whois.dump', 1);

 # Find information regarding the given IP address
 my @results = $wi_tree->find('146.102.1.1');

 # Find information regarding the given IP address
 my $abuse = $wi_tree->find_abuse('146.102.1.1');

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Value::Convertor;
use Value::IPv4ADDR;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant MAX_AGE => 259200;

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.6';
    $DEVEL   = 0;
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC ,STATIC]

  Usage     : my $instance = Whois::Service->new()
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Whois::Service reference on success, croaks on failure
  Throws    : Croaks, if invoked on object
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my $self = bless ({}, $class);
    return $self->_init(@_);
}

=item check() [PUBLIC]

  Usage     : my @warnings = $instance->check()
  Purpose   : Check the inner whois database for errors and warnings
  Arguments : None
  Returns   : ARRAY containing warnings
  Throws    : Dies, if not invoked on object instance

=cut

sub check()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my @warnings = ();
    foreach my $s (@{$self->{SERVICES}}) {
        my @w = $s->check();
        push(@warnings, @w) if @w;
    }
    return @warnings;
}

=item find($;$) [PUBLIC]

  Usage     : my $result = $instance->find('146.102.1.1');
  Purpose   : Search for given IPv4 address in the inner whois storage
  Arguments : MIXED $ipv4 - IPv4 address
  Returns   : ARRAY OF HASH REFERENCES containing range structures
  Throws    : Dies, if not invoked on object instance

=cut

sub find($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my @ranges = ();
    foreach my $s (@{$self->{SERVICES}}) {
        @ranges = $s->find(@_);
        last if @ranges;
    }
    return @ranges;
}

=item rfind($) [PUBLIC]

  Usage     : my $result = $instance->rfind('email@abuse.domain');
  Purpose   : Search for given email address in the inner whois storage
  Arguments : MIXED $email - email address [MANDATORY]
  Returns   : ARRAY OF HASH REFERENCES containing network structures
  Throws    : Dies, if not invoked on object instance

=cut

sub rfind($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my @ranges = ();
    foreach my $s (@{$self->{SERVICES}}) {
        @ranges = $s->rfind(@_);
        last if @ranges;
    }
    return @ranges;
}

=item find_abuses($) [PUBLIC]

  Usage     : my $result = $instance->find_abuses('146.102.1.1');
  Purpose   : Search for abuse contacts for given IP address in the inner whois storage
  Arguments : Same as find() method
  Returns   : STRING $abuse contact on SUCCESS, UNDEF on FAILURE
  Throws    : Dies, if not invoked on object instance
  See also  : find() method

=cut

sub find_abuses($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my $result;
    foreach my $s (@{$self->{SERVICES}}) {
        $result = $s->find_abuses(@_);
        last if $result;
    }
    return $result;
}

=item find_networks($) [PUBLIC]

  Usage     : my $result = $instance->find_networks('abuse@domain');
  Purpose   : Search for abuse contacts for given email address in the inner whois storage
  Arguments : Same as find() method
  Returns   : STRING $abuse contact on SUCCESS, UNDEF on FAILURE
  Throws    : Dies, if not invoked on object instance
  See also  : find() method

=cut

sub find_networks($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my $result;
    foreach my $s (@{$self->{SERVICES}}) {
        $result = $s->find_networks(@_);
        last if $result;
    }
    return $result;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init($) [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Whois::Service instance
#  Arguments : STRING $whois_file - name of the file containing whois dump
#              BOOL $force - parse the whois dump file, even if it is too old
#  Returns   : Whois::Service reference
#  Throws    : Dies, if not invoked on object instance
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init($;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my @services = @_;
    die "Whois services must be given as argument" unless @services;

    $self->{SERVICES} = \@services;

    return $self;
}

=pod

=back

=cut

1;
